'''
Created on 20.05.2014/ last change 29.1.18

@author: estevez schwarz/lamour
'''
import logging
from datetime import datetime
from numpy import  shape, empty, append,transpose,amax,eye
from scipy.linalg import pinv,svd
from algopy import  dot, zeros
from toolboxLA.svdinfo import CompCondSing ,OrthogonalProjectorOntoKer, OrthogonalBasisProjectorAlongIm, RankSigma
from NumericalException import LittleNumberOfTaylorCoefficients, SingularDAE
from algopy.utpm import UTPM
from toolboxAD.MyUTPM import MyUTPM

class TaylorDAE(object):
    '''
    TaylorDAE object has the attributes
        :attribute:
            * **example** - object of type "example" (see Implementation of an example)
            * **K** - Number of Taylor coefficients 0,...,K, i.e., we have K+1 coefficients
            * **tol** - absolute accuracy for rank determination
            * **info_level** controls the information output
                * 0 - no output\n
                * 1 ... 4 output increases
    '''
    
    def __init__(self,example,K,tol=None,info_level=2):
        self.example=example
        self.K=K
        self.tol=tol
        self.info_level=info_level
        if not logging.getLogger("array").handlers: # we use the property of every object having a boolean value True
            self.log_files()
        self.normDerivativeArray=0.0 # value to transport the value of norm(derivativeArray) from consistentValues.derivativeArray to consistentValues.consistentValues
       
    def log_files(self):
        """
        log_files provides a list which contains the Logger for the different files. 
        
        The following numerical submethods are logged:
            * 0 - computation of consistent initial values
            * 1 - Newton method (not active for InitDAE)
            * 2 - linear algebra results of derivative array computations
            * 3 - integration of IVPs (not active for InitDAE)
        """
        if self.info_level>3:
            print 'TaylorDAE.log_files: info_level greater then 3, reset to 3!'
            self.info_level=3
        if self.info_level>0:
            time=str(datetime.now())
            
            logfile_consi=self.example.example_name+'_consi_'+time[:-16]+".log"
            logfile_newton=self.example.example_name+'_newton_'+time[:-16]+".log"
            logfile_array=self.example.example_name+'_array_'+time[:-16]+".log"
            logfile_integration=self.example.example_name+'_integr_'+time[:-16]+".log"
            
            if self.info_level==3:
                mylevel=logging.DEBUG
            if self.info_level==2:
                mylevel=logging.INFO
            if self.info_level==1:
                mylevel=logging.WARNING
                
            logging.basicConfig(level=mylevel)
            
            consi_log = logging.getLogger("consi")
            newton_log = logging.getLogger("newton")
            array_log = logging.getLogger("array")
            integr_log = logging.getLogger("integr")
            
            consi_log.propagate=False
            newton_log.propagate=False
            array_log.propagate=False
            integr_log.propagate=False
            
            consiHandler=logging.FileHandler(logfile_consi)
            consi_log.addHandler(consiHandler)
            
            #newtonHandler=logging.FileHandler(logfile_newton)
            #newton_log.addHandler(newtonHandler)
            
            arrayHandler=logging.FileHandler(logfile_array)
            array_log.addHandler(arrayHandler)
            
            integrHandler=logging.FileHandler(logfile_integration)
            integr_log.addHandler(integrHandler)
            
            text='\nExample '+self.example.example_name+' started at '+time[:-7]#+' , info_level = '+str(mylevel)
            stars='    *****************************************************************'
            consi_log.warning(text+' number of Taylorcoeff. = '+str(self.K)+stars)
            #newton_log.warning(text+stars)
            array_log.warning(text+stars)
            integr_log.warning(text+stars)
        
    @staticmethod    
    def write_Values2logger(logger,info_level,print_level,message):
        """
        write_Values2logger stores messages depending of the info_level in the logger.\n
            :param:
                * **logger** - used logger 
                * **info_level** - if > 0 creates an output. The greater the more.
                * **print_level** - determines the level of the message.
                * **message** - string containing the message
        """
        if info_level>=print_level:
            if print_level>2:
                logger.debug(message)
            elif print_level==2:
                logger.info(message)
            elif print_level==1:
                logger.warning(message)
            else:
                print 'write_Values2logger: Wrong print_level =',print_level,' !'
                
    @staticmethod        
    def write_Values2file(name,message,dim,r,S1,Sr,Sr1,Send,tol,info_level):
        '''
        write_Values2file writes the values of the rank determination to a file
            :param: 
                * **name**
                * **message**
                * **dim**
                * **r** - see toolboxLA.svdinfo.CompCondSing
                * **S1**
                * **Sr**
                * **Sr1**
                * **Send**
                * **tol**
                * **info_level** - if >2 information printed on terminal
        '''
        
        string='  '+message+': noOfColumns= {:3d}  Rank= {:3d}'.format( dim,r)
        if dim > 2:
            # full rank case
            string +=' S(1)= {:9.6e} S(r)= {:9.6e}'.format(S1,Sr)
            if r==dim:
                string +=' Full rank.'
            elif r+1==dim:
                string += ' S(end)= {:9.6e}'.format(Send)
            else:
                string += ' S(r+1)= {:9.6e} S(end)= {:9.6e}'.format(Sr1,Send)
        else:
            if dim==1:
                string +=' S(1)= {:9.6e}'.format(S1)
                if r==dim:
                    string += ' Full rank.'
            else:
                if r==dim:
                    string +=' S(1)= {:9.6e} S(r)= {:9.6e} Full rank.'.format(S1,Sr)
                else:
                    string +=' S(1)= {:9.6e} S(end)= {:9.6e}'.format(S1,Sr1)
                    
        form= 'None' if type(tol)==type(None) else  '{:9.6e}'
        string += (' tol = '+form).format(tol)
               
        logger = logging.getLogger(name)
        logger.debug(string)
        if info_level > 2:
            print string
        
    def Jacobian_matrices_ADB(self,ti,zt):
        ''' 
        The method *Jacobian_matrices_ADB* computes the Jacobian matrices
            :param:  
                * **ti** - Taylor object
                * **zt** - MyUTPM object
            :return: 
                * **A** = df/dx'(zt,ti)
                * **D** = dd/dx(zt,ti)
                * **B** = df/dx(zt,ti)
        '''
        f=self.example.f
        d=self.example.d
        #print f,d
        (cgD,cgAB,f_Taylor)=zt.tracing_A_D_B(d,f,ti)
        (D,A,B)=zt.A_D_B_new(cgD, cgAB, d, f, ti)  
        #print "D,A,B\n",D,'\n\n',A,'\n\n',B
        
        return D,A,B,f_Taylor
    
    def Jacobian_matrices_AB(self,ti,zt):
        ''' 
        The method *Jacobian_matrices_AB* computes the Jacobian matrices 
            :param:  
                * **ti** - Taylor object
                * **zt** - MyUTPM object
            :return: 
                * **A** = df/dx'(zt,ti)
                * **B** = df/dx(zt,ti)
                * **f_Taylor** - derivative array of f (see MyUTPM.tracing_A_B)
        '''
        f=self.example.f
        cgAB,f_Taylor =zt.tracing_A_B(f,ti)
        (A,B)=zt.A_B(cgAB, f, ti)  
        #print "Jacobian_matrices_AB: A,B\n\n",A,"\n\n",B
        return A,B,f_Taylor
    
    def Jacobian_matrix_u(self,zt):
        ''' 
        The method *Jacobian_matrix_u* computes the Jacobian matrix of u(x)
            :param:  
                * **zt** - MyUTPM object
            :return: 
                * **u_x** = du/dx(zt)
        '''
        x = UTPM.init_jacobian(zt)
        ux = self.example.u(x)
        u_x = UTPM.extract_jacobian(ux)
        #print 'Jacobian_matrix u_x: ',u_x
        return u_x
    
    
    def Jacobian_forCoefficentMatrix(self,tti,zt):
        ''' 
        The method *Jacobian_forCoefficentMatrix* computes the Jacobian matrices Atilde and Btilde.
            :param:  
                * **tti** - Taylor object of timepoint
                * **zt** - MyUTPM object of z
            :return:
                * The output depends of the structure of the DAE. For 
                    * type_of_DAE = **'proper'**        
                        * Atilde= A.D, Btilde= A D'+ B with f( d'(x,t),x,t) and
                            A = df/dx'(zt,ti),
                            D = dd/dx(zt,ti),
                            B = df/dx(zt,ti)
                    * type_of_DAE = **'standard'**      
                        * Atilde= A, Btilde= B  with f( x',x,t)
                    * type_of_DAE = **'linear'** 
                        * Atilde= eval_A(t), Btilde= eval_B(t)  with f( x',x,t) = eval_A x'+eval_B x-q
            :raised:
                * **Exception** - unknown type of DAE
        '''
        
        if self.example.type_of_DAE == 'proper':
            (D,A,B,f_Taylor)=self.Jacobian_matrices_ADB(tti,zt)
            DD=MyUTPM(D.data)
            # Computing of S= A D'+B and G0=A D       
            Dprime=DD.diff() # reduction of number of Taylor coefficents
            Btilde=dot(A,Dprime)+B
            Atilde=dot(A,DD)
            
        elif self.example.type_of_DAE == 'standard':
            (Atilde,Btilde,f_Taylor)=self.Jacobian_matrices_AB(tti,zt) 
            
        elif self.example.type_of_DAE == 'linear':
            Atilde= self.example.eval_A(tti)
            Btilde= self.example.eval_B(tti)
            z=zt.diff()
            f_Taylor=self.example.f(z,zt,tti).data
            
        else:
            raise Exception("TaylorDAE.Jacobian_forCoefficentMatrix: No such type_of_DAE (proper/standard/linear)")
        
        #print 'DAE_TaylorDAE.Jacobian_forCoefficentMatrix: A \n\n',Atilde,'\n\nB \n\n',Btilde
        if self.info_level>0:
            logging.getLogger("array").info('\nJacobian_forCoefficentMatrix: A =\n\n'+str(Atilde.data[0,0])+'\n\n'+str(Atilde.data[1,0])\
                                  +'\n\nJacobian_forCoefficentMatrix: B= \n\n'+str(Btilde.data[0,0])+'\n\n'+str(Btilde.data[1,0]))
        return Atilde,Btilde,f_Taylor
    
    def coefficent_matrix(self,tti,xt):
        '''
        coefficient matrix of Taylor method
            :param:
                * **tti** - Taylor representation of ti (time point)
                * **xt** - Taylor coefficients of x(ti)
            :return: 
                * **Coeff_Matrix** - see structure below
                * **block_dimension** - n
        structure of Coeff_Matrix for constant standard DAE Ax'+Bx = q
            +---+---+---+---+---+---------------------------------------------+
            | A |   |   |   |   |                                             |
            +---+---+---+---+---+---------------------------------------------+
            | B | A |   |   |   |    n(K+1) x n(K+1)                          |
            +---+---+---+---+---+---------------------------------------------+
            |   | B | A |   |   |   self.K+1 - number of Taylor coefficients  |
            +---+---+---+---+---+---------------------------------------------+
            |   |   |...|...|   |   A, B  nxn matrices                        |
            +---+---+---+---+---+---------------------------------------------+   
            |   |   |   | B | A |                                             |
            +---+---+---+---+---+---------------------------------------------+
        '''
         
        (G0,S,f_Taylor)=self.Jacobian_forCoefficentMatrix(tti,xt)
        #print 'coefficent_matrix: s.data.shape= ',S.data.shape
        dim = shape(G0.data[0,0])
        dimension=dim[0]*(self.K)
        #print 'coefficent_matrix: dimension, self.K = ',dimension,self.K
        Coeff_Matrix=zeros((dimension,dimension)) 
         
        # filling of the matrix
        for i in xrange(self.K): # index for matrix dimension    #######################################
            shift=i*dim[0]
            shift_c=0
            for j in xrange(self.K-i-1): # index for Taylor coefficient
                ds=dim[0]+shift # row index
                dspdim=ds+dim[0]
               
                Coeff_Matrix[ds+shift_c:dspdim+shift_c,shift_c:shift_c+dim[0]]=S.data[i,0]
                if i==0:
                    Coeff_Matrix[ds+shift_c:dspdim+shift_c,ds+shift_c:dspdim+shift_c]=(j+1)*G0.data[0,0]
                else:
                    Coeff_Matrix[ds+shift_c:dspdim+shift_c,dim[0]+shift_c:2*dim[0]+shift_c] +=(j+1)*G0.data[i,0]
                shift_c=shift_c+dim[0]
        Coeff_Matrix[0:dim[0],0:dim[0]]=G0.data[0,0]
        #print 'Coeff_Matrix \n\n',Coeff_Matrix
        return Coeff_Matrix,dim[0],f_Taylor
    
    def coefficent_matrix_optimization(self,tti,xt):
        '''
        coefficient matrix of Taylor method
            :param:
                * **tti** - Taylor representation of ti (time point)
                * **xt** - Taylor coefficients of x(ti)
            :return: 
                * **Coeff_Matrix** - 
                * **block_dimension** -
        structure of Coeff_Matrix for constant standard DAE Ax'+Bx = q
            +---+---+---+---+---+---------------------------------------------+
            | B | A |   |   |   |    n*K x n(K+1)                             |
            +---+---+---+---+---+---------------------------------------------+
            |   | B | A |   |   |    self.K+1 - number of Taylor coefficients |
            +---+---+---+---+---+---------------------------------------------+
            |   |   |...|...|   |    A, B  nxn matrices                       |
            +---+---+---+---+---+---------------------------------------------+   
            |   |   |   | B | A |                                             |
            +---+---+---+---+---+---------------------------------------------+
        '''
        (G0,S,f_Taylor)=self.Jacobian_forCoefficentMatrix(tti,xt)
        dim = shape(G0.data[0,0])
        dimension=dim[0]*(self.K-1)
        Coeff_Matrix=zeros((dimension,dimension+dim[0])) 
        
        # filling of the matrix
        for i in xrange(self.K-1): # index for matrix dimension    #######################################
            shift=i*dim[0]
                
            shift_c=0
            for j in xrange(self.K-i-1): # index for Taylor coefficient
                ds=shift # row index
                dspdim=ds+dim[0]
                    
                Coeff_Matrix[ds+shift_c:dspdim+shift_c,shift_c:shift_c+dim[0]]=S.data[i,0]
                if i==0:
                    Coeff_Matrix[ds+shift_c:dspdim+shift_c,dim[0]+ds+shift_c:dim[0]+dspdim+shift_c]=(j+1)*G0.data[0,0]
                else:
                    Coeff_Matrix[ds+shift_c:dspdim+shift_c,dim[0]+shift_c:2*dim[0]+shift_c] +=(j+1)*G0.data[i,0]
                shift_c=shift_c+dim[0]
        #print 'Coeff_Matrix \n\n',Coeff_Matrix
        return Coeff_Matrix,dim[0],f_Taylor

    
    def CheckOmegaTaylorAndProjectors(self,Coeff_Matrix,n):
        '''
        CheckOmegaTaylorAndProjectors checks the index of the DAE and computes the projectors T_j.
            :param:
                * **Coeff_Matrix** -  Jacobian matrix of the derivative array with additional A at the first block row 
                * **n** - Number of equations of the original DAE 
            :return: 
                * **T** - Projectors T_j, all in one matrix 
                * **index** - Differentiation index of the DAE       
        '''
         
        tol=self.tol
        info_level=self.info_level
        r1old=-1
        no_constraints=False
        index=-1
        #storage of projectors
        T=empty([n,n,1])
        # loop to find the index
        for i in xrange(self.K+1): 
            if i==0:
                A0=Coeff_Matrix[:n,:n]
                (T, rankA, S1, Sr, Sr1, Send,used_tol) = OrthogonalProjectorOntoKer(A0,tol=tol)
                stars='*'*86+'\n'
                message='\n'+stars+'CheckOmegaTaylorAndProjectors: Projector based decomposition of the derivative array.\n'+stars
                self.write_Values2logger(logging.getLogger("array"),info_level,3,message)
                if info_level>2:
                    print message
                self.write_Values2file("array", "A             ", n,rankA, S1, Sr, Sr1, Send,used_tol,info_level)
                if rankA == n:
                    index=0
                    if self.info_level>0:
                        logging.getLogger("array").info('\nCheckOmegaTaylorAndProjectors: Index = '+str(i)+'\n')
                    break
            else:
                GR=Coeff_Matrix[n:(i+1)*n,n:(i+1)*n]
                GL=Coeff_Matrix[n:(i+1)*n,0:n]
                (WGRbase,rR,S1,Sr,Sr1,Send,used_tol)=OrthogonalBasisProjectorAlongIm(GR,tol=tol)
                self.write_Values2file("array","Computation WR",i*n,rR,S1,Sr,Sr1,Send,used_tol,info_level)
                H1=dot(WGRbase,GL) # H1 corresponds to N in the articles
                
                #dim_WGRbase=shape(WGRbase)
                dim_H1=shape(H1)
                
                [QH1,r1,S1,Sr,Sr1,Send,used_tol]=OrthogonalProjectorOntoKer(H1,tol=tol)
                #print 'CheckOmegaTaylorAndProjectors: shape(H1),dim_WGRbase r1,S1,Sr,Sr1,Send = ',shape(H1),dim_WGRbase,r1,S1,Sr,Sr1,Send
                #self.write_Values2file("array","WR*GL         ",min(dim_WGRbase[0],dim_WGRbase[1]),r1,S1,Sr,Sr1,Send,used_tol)
                #self.print_SVDValues(info_level,"WR*GL         ",min(dim_WGRbase[0],dim_WGRbase[1]),r1,S1,Sr,Sr1,Send,used_tol)
                self.write_Values2file("array","WR*GL         ",min(dim_H1[0],dim_H1[1]),r1,S1,Sr,Sr1,Send,used_tol,info_level)
                if r1==r1old:
                    print('CheckOmegaTaylorAndProjectors: No new constraints after differentiation')
                    if self.info_level>0:
                        logging.getLogger("array").info('\n########################\n- NO NEW CONSTRAINTS AFTER DIFFERENTIATION -\n########################\n')
                    no_constraints=True
                    break
                r1old=r1
         
                H2=append(Coeff_Matrix[0:n,0:n],H1,axis=0)
                T=append(T,empty((n,n)),axis=0)
                [T[n*i:,:],r2,S1,Sr,Sr1,Send,used_tol]=OrthogonalProjectorOntoKer(H2,tol=tol)

                self.write_Values2file("array","Computation T ",n,r2,S1,Sr,Sr1,Send,used_tol,info_level)
                if i==1 and self.info_level>2:
                    print "Projector P: \n", eye(n)-T[0:n,0:n],'\n'
                
                
                ######################### Test von Diana zur Berechnung der Projektoren Vi               
                     
                GLQ=dot(GL,T[0:n,0:n])
                GLQ_GR=append(GLQ,GR,axis=1) 
                #print 'CheckOmegaTaylorAndProjectors: size (GLQ,GR)',shape(GLQ_GR),GLQ_GR
                (W_GLQ_GRbase,rR,S1,Sr,Sr1,Send,used_tol)=OrthogonalBasisProjectorAlongIm(GLQ_GR,tol=tol)
                                    
                H3d=dot(W_GLQ_GRbase,GL)
                #print 'CheckOmegaTaylorAndProjectors: size (W GL)',shape(H3d),W_GLQ_GRbase
                H3=append(T[0:n,0:n],H3d,axis=0)
                #print 'CheckOmegaTaylorAndProjectors: size (H3)',shape(H3),H3
                [Vi,r3,S13,Sr3,Sr13,Send3,used_tol3]=OrthogonalProjectorOntoKer(H3,tol=tol)
                if self.info_level>2:
                    print "Projector V_",i,": \n", Vi,'\n'
                ####################################################
                if r2==n:
                    if self.info_level>0:
                        logging.getLogger("array").info('\nCheckOmegaTaylorAndProjectors: Index = '+str(i)+'\n')
                    index=i
                    break
                # Remark: Number of Taylor coefficients too small for index determination
                if i+1==self.K:
                    print "CheckOmegaTaylorAndProjectors: Number of Taylor coefficients = ",self.K,\
                        " too small for index determination"
                    logging.getLogger("array").info("CheckOmegaTaylorAndProjectors: Number of Taylor coefficients = {0:d} too small for index determination".format(self.K))
                    raise LittleNumberOfTaylorCoefficients(self.K)
                     
        if  no_constraints:
            print "CheckOmegaTaylorAndProjectors: No more constraints to differentiate. Looks like a singularity of the DAE."
            if self.info_level>0:
                logging.getLogger("array").info("CheckOmegaTaylorAndProjectors: No more constraints to differentiate.\n Looks like a singularity of the DAE. Deep of the loop = {0:d}.".format(i))
            raise SingularDAE("Singular DAE")
        if self.info_level>0:
            for i in xrange(T.shape[0]/n-1):
                imn=i*n
                logging.getLogger("array").debug('\nCheckOmegaTaylorAndProjectors: T_'+str(i)+' =\n\n'+str(T[imn:imn+n,:])+'\n')
         
        return T,index
    
    def IndexAndCondition(self, tti, zt):
        """
        IndexAndCondition computes the index and the condition number of the DAE.
            :param: 
                * **tti** - Taylor object of ti
                * **zt** -  Taylor object of z(t)
            :return:
                * **index** - index of the DAE
                * **condM** - condition number of index determination
                * **s** - level of s-fullness
        """
        if self.info_level>0:
            logging.getLogger("array").info('IndexAndCondition: time = '+str(tti.data[0,0]))
        tol=self.tol
         
        [CM, n,dummy] = self.coefficent_matrix(tti, zt)
        # Check of row regularity
        k_level_sing=-1
        for k_level in xrange(2,self.K+1):
            [Ux,Sigma,VTx]=svd(CM[n:k_level*n,:])
            #[Ux,Sigmax,VTx]=svd(CM)
            noOfRows=len(Sigma)
            rank,used_tol=RankSigma(Sigma, tol)
            if rank != noOfRows:
                print 'IndexAndCondition: number of rows = {:6d} rowrank = {:6d}'.format(noOfRows,rank) 
                k_level_sing=k_level-1
                break
        # check of s-fullness
        s=self.Check_s_full(CM,n)
        #print 'IndexAndCondition: s-full s= ',s
        TProj, index = self.CheckOmegaTaylorAndProjectors(CM, n)
        #print 'IndexAndCondition: index = {:4d}'.format(index)
        if k_level_sing > -1:
            if k_level_sing <= index or ( index==0 and k_level_sing ==1):
                #print 'IndexAndCondition: k_level_sing  index ',k_level_sing,  index
                raise SingularDAE("IndexAndCondition: Matrixpencil Singular - Rowrank of G^[{:4d}] = {:4d} .ne. number of rows = {:4d}. Sigma(end) = {:9.6e}, used_tol = {:9.6e}"\
                                  .format(k_level_sing,rank,noOfRows,Sigma[-1],used_tol))
            else:
                #print 'IndexAndCondition: k_level_sing  index ',k_level_sing,  index
                print 'IndexAndCondition: Warning - Rowrank of G^[{:4d}] = {:4d} .ne. number of rows = {:4d}. Condition number problems.'.format(k_level_sing,rank,noOfRows)
        # calculation of condition number
        indexA = index
        if index == 0:
            indexA = 1
        (condM, rankM, S0, SrankM, Srp1, Send, used_tol) = CompCondSing(CM[0:(indexA + 1) * n, 0:(indexA + 1) * n],tol=tol)
         
        self.write_Values2file("array", "Condition CM  ", (indexA + 1) * n, rankM, S0, SrankM, Srp1, Send, used_tol,self.info_level)
        return index,condM,s
    
    def Check_s_full(self,CoeffMatrix,n):
        """
        Check_s_full computes the value s, which characterizes the accuracy of the result.
        """
        G0=CoeffMatrix[:n,:n]
        Gm=pinv(G0)
        P0=dot(Gm,G0)
        CoeffMatrix[:n,:n]=P0
        #dim=shape(CoeffMatrix)
        [U,Sigma,VT]=svd(CoeffMatrix)
        #print 's-full: Sigma ',Sigma
        tol=self.tol
        rank,used_tol=RankSigma(Sigma,tol=tol)
        D=CoeffMatrix.shape[0]/n
        if self.info_level>2:
            print '\nCheck_s_full: rank = ',rank, 'used tolerance = ',used_tol,
        V=transpose(VT)
        V13_23=V[:rank,rank:]
        
        cond=Sigma[0]/Sigma[rank-1]
        if self.info_level>2:
            print 'condition number of B^k = {0:9.2e}'.format(cond)

        if rank==CoeffMatrix.shape[0]:
            s=D-1 # The last Taylorkoefficients are in the index 0 case correct. Therefor increase s by 1 in this case if you use it.
        else:
            max_V=0.0
            jp1=1
            eps_cond=2.220446049250313e-16*cond
            if self.info_level>2:
                stars='*'*57+'\n'
                print '\n\n',stars,'Check_s_full: Determination of s-fullness considering V\n',stars
            while max_V < eps_cond and (jp1*n<=rank):
                
                j=jp1-1
                max_V=amax(abs(V13_23[j*n:jp1*n,:]))
                jp1+=1
                if self.info_level>2:
                    print 'Check_s_full:  j = ',j, 'max_V = ',max_V,'eps*cond = ',eps_cond
                
            if max_V < eps_cond:
                s=jp1-1
            else:        
                s=j 
        return s

    def f_taylor(self,x,t):
        '''
        f_taylor computes the Taylor coefficients of f depending of the DAE type
            :param: 
                * **x** - Taylor object of x(t)
                * **t** - Taylor object of t
            :return:
                * **f** - Taylor expansion of f
            :raised:
                * **Exception** - unknown type of DAE
        '''
        if self.example.type_of_DAE == 'proper':
            dxt=self.example.d(x,t)
            dxt_prime=dxt.diff()
            f_Taylor=self.example.f(dxt_prime,x,t).data
            
        elif self.example.type_of_DAE == 'standard':
            xt=x.diff()
            f_Taylor=self.example.f(xt,x,t).data 
        elif self.example.type_of_DAE == 'linear':
            xt=x.diff()
            f_Taylor=self.example.f(xt,x,t).data #noch testen
            
        else:
            raise Exception("TaylorDAE.f_taylor: No such type_of_DAE (proper/standard/linear)")
        return f_Taylor
